{   obcc, a small oberoncompiler for the C16x-family
      based on Oberon-0 by N. Wirth

    Copyright 2011 Guido B. (user "guido-b" at http://www.mikrocontroller.net)

    obcp.pas, the parser, is part of obcc.
    
    obcc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    obcc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with obcc.  If not, see <http://www.gnu.org/licenses/>.                }




unit obcp; { Der Parser }

interface

PROCEDURE Compile(fname: String);

var verstr: STRING;

implementation
uses obcs, obcg;

CONST WordSize = 2;
VAR sym: INTEGER; 
    topScope, universe: g_Object; (* linked lists, end with guard *)
    guard: g_Object; 
    ProcPar: BOOLEAN;
    
	PROCEDURE NewObj(VAR obj: g_Object; class: INTEGER);
	  VAR pnew, x: g_Object;
	BEGIN x := topScope; guard^.name := id;
	  WHILE x^.next^.name <> id DO x := x^.next;
	  IF x^.next = guard THEN BEGIN
	    NEW(pnew); pnew^.name := id; pnew^.class := class; 
	    pnew^.IsPara := FALSE; pnew^.next := guard;
	    x^.next := pnew; obj := pnew; END
	  ELSE BEGIN obj := x^.next; Mark('mult def'); END;
	END {NewObj};
	
	PROCEDURE find(VAR obj: g_Object);
	  VAR s, x: g_Object; ok: BOOLEAN;
	BEGIN s := topScope; guard^.name := id; ok := FALSE;
	  REPEAT x := s^.next;
	    WHILE x^.name <> id DO x := x^.next;
	    IF x <> guard THEN BEGIN obj := x; ok := TRUE; END
	    ELSE IF s = universe THEN BEGIN 
	      obj := x; Mark('undef: ' + Id); ok := TRUE; END 
	      ELSE s := s^.dsc;
	  until ok;  
	END {find};
	
	PROCEDURE FindField(VAR obj: g_Object; list: g_Object);	
	BEGIN guard^.name := id;
	  WHILE list^.name <> id DO list := list^.next;
	  obj := list;
	END {FindField};
	
	Function IsParam(obj: g_Object): BOOLEAN;
	BEGIN 
	  IF (obj^.class = g_Par) OR ((obj^.class = g_Var) AND (obj^.IsPara)) then
	    IsParam := TRUE
	  ELSE IsParam := FALSE;
	END {IsParam};

	PROCEDURE OpenScope;
	  VAR s: g_Object;
	BEGIN NEW(s); s^.class := g_Head; s^.dsc := topScope; s^.next := guard; topScope := s;
	END {OpenScope};
	
	PROCEDURE CloseScope;
	BEGIN topScope := topScope^.dsc;
	END {CloseScope};
	
	PROCEDURE CheckBool(VAR x: g_Item);
	BEGIN
	  IF x.i_type^.form <> g_Boolean THEN Mark('not boolean');
	END;
	
  (* -------------------- Parser ---------------------*)	
  
  	PROCEDURE expression(VAR x: g_Item); FORWARD;
  
  	PROCEDURE selector(VAR x: G_Item);
  	  VAR y: G_Item; obj: G_Object;
	BEGIN
	  WHILE (sym = lbrak_) OR (sym = period_) DO BEGIN
	    IF sym = lbrak_ THEN BEGIN
	      Get(sym); expression(y);
	      IF x.i_type^.form = g_Array THEN Index(x, y) ELSE Mark('not an array');
	      IF sym = rbrak_ THEN Get(sym) ELSE Mark(']?');END
	    ELSE BEGIN { period_ }
	      Get(sym);
	      IF sym = ident_ THEN 
	        IF x.i_type^.form = g_Record THEN BEGIN
	          FindField(obj, x.i_type^.fields); Get(sym);
	          IF obj <> guard THEN BEGIN 
	            Field(x, obj); x.i_type := obj^.o_type; 
	          END
	          ELSE Mark('undef'); end
	        ELSE BEGIN Mark('not a record'); END
	      ELSE Mark('ident?');
	    END;
	  END;
	END {selector};
	
	PROCEDURE factor(VAR x: g_Item);
	  VAR obj: g_Object;
	BEGIN (*sync*)
	  IF sym < lparen_ THEN BEGIN
	    Mark('ident?');
	    REPEAT Get(sym) UNTIL sym >= lparen_; END ;
	  case sym of
	    ident_:  BEGIN
	      find(obj); Get(sym); MakeItem(x, obj); selector(x); END;
	    number_:  BEGIN
	      MakeConstItem(x, intType, val); Get(sym); END;
	    str_: BEGIN
	      MakeStringItem(x, StringType, id); Get(sym); END;
	    lparen_: BEGIN
	      Get(sym); expression(x); 
	      IF sym = rparen_ THEN Get(sym) ELSE Mark(')?'); END;
	    not_ : BEGIN
	      Get(sym); factor(x); Op1(not_, x); end
	     ELSE BEGIN Mark('factor?'); MakeItem(x, guard); end;
	  END;
	END {Factor};
	
	
	PROCEDURE term(VAR x: g_Item);
	  VAR y: g_Item; op: INTEGER;
	BEGIN factor(x);
	  WHILE (sym >= times_) AND (sym <= and_) DO BEGIN
	    op := sym; Get(sym);
	    IF op = and_ THEN BEGIN
	      Op1(op, x);
	      IF sym <> lparen_ THEN Mark('( ?'); END;
	    factor(y); Op2(op, x, y);
	  END;
	END {term};
	
	PROCEDURE SimpleExpression(VAR x: g_Item);
	  VAR y: g_Item; op: INTEGER;
	BEGIN
	  IF sym = plus_ THEN BEGIN  Get(sym); term(x); end
	  ELSE IF sym = minus_ THEN BEGIN 
	    Get(sym); term(x); Op1(minus_, x); end
	  ELSE term(x);
	  { Das ist nicht optimal, da keine boolschen Verknpfungen fr
	    Prozedurparameter damit mglich sind: TODO }
	  IF (x.i_type^.form = g_Boolean) AND ((x.mode = g_Var) OR (x.mode = g_Reg)) 
	      AND (sym IN [then_, do_, semicolon_, rparen_]) AND (NOT ProcPar) THEN BEGIN
	    TestTrue(x);
	  END;
	  WHILE (sym >= plus_) AND (sym <= or_) DO BEGIN
	    op := sym; Get(sym);
	    IF op = or_ THEN BEGIN 
	      Op1(op, x);
	      IF sym <> lparen_ THEN Mark ('( ?'); END;
	    term(y); Op2(op, x, y);
	  END;
	END {SimpleExpression};
  
  	PROCEDURE expression(VAR x: g_Item);
  	  VAR y: g_Item; op: INTEGER;
	BEGIN SimpleExpression(x);
	  IF (sym >= eql_) AND (sym <= gtr_) THEN BEGIN
	    op := sym; Get(sym); SimpleExpression(y);
	     Relation(op, x, y); END;
  	END {expression};
	
	PROCEDURE parameter(VAR fp: g_Object; sc: BOOLEAN);
	  VAR x: g_Item;
	BEGIN ProcPar := TRUE; expression(x); ProcPar := FALSE;
	  IF IsParam(fp) THEN BEGIN 
	    g_Parameter(x, fp^.o_type, fp^.class, sc); fp := fp^.next; 
	    ParCount := ParCount + 1; end
	  ELSE Mark('too many parameters');
	END {parameter};
	
	
	PROCEDURE StatSequence;
	  VAR par, obj: g_Object; x, y: g_Item; var ok, done, lastforl: BOOLEAN;
	      label1, label2: LONGINT; smallcall: BOOLEAN;
	
		PROCEDURE param(VAR x: g_Item);
		BEGIN
		  IF sym = lparen_ THEN Get(sym) ELSE Mark('(?');
		  //IF sym = str_ THEN BEGIN
		   // x.mode := g_String; x.tok := id; Get(sym); END
		  //ELSE 
		  ProcPar := TRUE;
		  expression(x); ProcPar := FALSE;
		  IF sym = rparen_ THEN Get(sym) ELSE Mark(')?');
		END {param};
		
		PROCEDURE wparams(VAR x: g_Item); { Parameter von Write und Writeln }
		VAR ok : BOOLEAN;
		BEGIN ok := FALSE; Get(sym);
		  REPEAT
		      ProcPar := TRUE; expression(x); ProcPar := FALSE;
		      IF (x.i_type^.form = g_Array)  THEN
		        Mark('bad writeparameter!')  ;
		      IF sym = colon_ THEN BEGIN { Formatanweisung }
		        Get(sym);
		        IF (sym = ident_) AND (id = 'h') THEN BEGIN
		          PutWrite(x, 1); { Hexadezimal }
		          Get(sym); END
		        ELSE mark('format?');
		      END
		      ELSE BEGIN
		        PutWrite(x,0);
		      END;
		      IF sym = comma_ THEN Get(sym)
	                ELSE IF sym = rparen_ THEN BEGIN 
	                  Get(sym); ok := TRUE; end
	                  ELSE IF sym >= semicolon_ THEN ok := TRUE
	                    ELSE Mark(') or , ?')
		  UNTIL ok;
		END; {wparams}
	  
	BEGIN (* StatSequence *)
	done := FALSE;
	REPEAT (*sync*) obj := guard;
    	  IF sym < ident_ THEN BEGIN
    	    Mark('statement?');
    	    REPEAT Get(sym) UNTIL sym >= ident_; end;
    	  CASE sym of
	    ident_: BEGIN
	      find(obj); MakeItem(x, obj); Get(sym); selector(x); 
	      Case sym of
	        becomes_: BEGIN
	          Get(sym); expression(y); Store(x, y); 
	        END;
	        eql_:  BEGIN
	          Mark(':= ?'); Get(sym); expression(y); END;
	        ELSE Begin
	          IF (x.mode = g_Proc) OR (x.mode = g_ProcR) THEN BEGIN
	            IF Isint THEN 
	              Mark('No procedure-call in interrupt!')
	            ELSE BEGIN
	              smallcall := smallproc;
	              IF x.mode = g_ProcR THEN smallproc := TRUE
	              ELSE smallproc := FALSE;
	              par := obj^.dsc;
	              g_OpenProc(obj^.val, smallproc);
	              ParCount := 0;
	              IF sym = lparen_ THEN BEGIN
	                Get(sym);
	                IF sym = rparen_ THEN Get(sym)
	                ELSE BEGIN
	                  ok := FALSE;
	                  REPEAT
	                    parameter(par, smallcall); 
	                    IF sym = comma_ THEN Get(sym)
	                    ELSE IF sym = rparen_ THEN BEGIN 
	                      Get(sym); ok := TRUE; end
	                      ELSE IF sym >= semicolon_ THEN ok := TRUE
	                        ELSE Mark(') or , ?')
	                  UNTIL ok;
                        END;
                      END;
                    { obj^.val ist seine locblksize, eigentlich mssten durch
                      asl auch forward-calls funktionieren, ein test darauf
                      ist hier wohl nicht mehr mglich}
                      IF obj^.val < 0 THEN Mark('forward call')
                      ELSE IF NOT IsParam(par) THEN Call(obj)
                        ELSE Mark('too few parameters');
                      smallproc := smallcall;
                    END;
                  END
                  ELSE IF x.mode = g_SProc THEN BEGIN
  		    IF (obj^.val = 2) OR (obj^.val = 4) THEN BEGIN {Write oder Writeln }
  		      IF sym = lparen_ THEN BEGIN
  		        OpenWrite; wparams(y); 
  		        IF obj^.val = 4 THEN CloseWrite(TRUE)
  		        ELSE CloseWrite(FALSE);
  		      END
  		      ELSE IF obj^.val = 2 THEN mark(' (?!')
  		        ELSE IOCall(x,y); { Writeln ohne Parameter }
  		    END
  		    ELSE Begin
                      IF obj^.val <= 3 THEN param(y); { Read oder WriteHex }
                      IOCall(x, y); END;
                  END
                  ELSE IF obj^.class = g_Typ THEN Mark('illegal assignment?')
                  ELSE Mark('statement?!'); 
                END;
              END;
            END;
            for_: BEGIN Get(sym);
              label1 := GetLabel; label2 := GetLabel; 
              lastforl := forloop; forloop := TRUE;
                IF sym = ident_ THEN BEGIN 
                  find(obj); MakeItem(x, obj); selector(x); Get(sym);
                  IF sym = becomes_ THEN BEGIN
                    Get(sym); expression(y); store(x,y); 
                    IF sym = to_ THEN BEGIN
                      Get(sym); expression(y); 
                      load(x);
                      IF (y.mode <> g_Const)  THEN load(y);                     
                      NewLabel(label1,'start for');                                            
                      relation(leq_, x, y); CJump(label2); load(x);                      
                      IF sym = do_ THEN BEGIN
                        Get(sym); StatSequence; g_Inc(x);
                        BJump(label1); NewLabel(label2, 'end for');
                        IF sym = end_ then Get(sym) ELSE Mark('END?');
                        IF y.mode <> g_Const THEN retreg(y);
                        StoreLv(x); forloop := lastforl;
                      END
                      ELSE Mark('DO?');
                    END
                    ELSE Mark('TO?');
                  END
                  ELSE Mark(' assignement?')
                
                END
                ELSE Mark(' FOR IDENT?');
            
            END;
            if_: BEGIN
             label1 := GetLabel; label2 := GetLabel;
              Get(sym); expression(x); 
              CJump(label1);
              IF sym = then_ THEN Get(sym) ELSE Mark('THEN?');
              StatSequence; 
              IF (sym = elsif_) or (sym = else_) then 
                BJump(label2);
              NewLabel(label1, ''); 
              WHILE sym = elsif_ DO BEGIN
                Get(sym); Label1 := GetLabel;
                 expression(x); 
                 CJump(label1);
                IF sym = then_ THEN Get(sym) ELSE Mark('THEN?');
                StatSequence; BJump(label2); NewLabel(label1, '');
              END;
              IF sym = else_ THEN BEGIN
                Get(sym); ; StatSequence; END;
              IF sym = end_ THEN Get(sym) ELSE Mark('END?');
              NewLabel(label2, '');
            END;
            repeat_: BEGIN Get(sym);
              label1 := GetLabel; NewLabel(label1,'');
              StatSequence;
              IF sym = until_ THEN BEGIN 
              	 Get(sym); expression(x); 
              	 CJump(label1); Get(sym); END
              ELSE Mark('until?');
            END;
            verb_: BEGIN
              putverb; Get(sym);
            END;            
            while_: BEGIN
              Get(sym); 
              label1 := getlabel; { Returnlink to start }
              NewLabel(label1, '');
              label2 := getlabel; { end of while-loop }
              expression(x); 
              CJump(label2);
              IF sym = do_ THEN Get(sym) ELSE Mark('DO?');
              StatSequence; BJump(label1); NewLabel(label2, '');
              IF sym = end_ THEN Get(sym) ELSE Mark('END?');
            END;
            semicolon_: Get(sym);
            (semicolon_ +1) .. (if_ -1) : done := TRUE;
           ELSE IF (sym >= array_) THEN done := TRUE
           	ELSE  Mark('; ?');
          end; 	
	until done;
	END {StatSequence};
	
	PROCEDURE IdentList(class: INTEGER; VAR first: g_Object);
	  VAR obj: g_Object;
	BEGIN
	  IF sym = ident_ THEN BEGIN
	    NewObj(first, class); Get(sym);
	    WHILE sym = comma_ DO BEGIN
	      Get(sym);
	      IF sym = ident_ THEN BEGIN
	        NewObj(obj, class); Get(sym); END
	      ELSE Mark('ident?');
	    END;
	  END;
	  IF sym = colon_ THEN Get(sym) ELSE Mark(':?');
	END {IdentList};
	
	PROCEDURE TType(VAR typet: g_Type);
	  VAR obj, first: g_Object; x: g_Item; tp: g_Type; ok:BOOLEAN;
	
	BEGIN typet := intType; (*sync*)	
	  IF (sym <> ident_) AND (sym < array_) THEN BEGIN
	    Mark('type?');	
	    REPEAT Get(sym) UNTIL (sym = ident_) OR (sym >= array_);
	  end;
	  CASE sym OF
	    ident_: BEGIN find(obj); Get(sym);
	             IF obj^.class = g_Typ THEN typet := obj^.o_type ELSE Mark('type?'); END;
	    array_: BEGIN IF (smallproc) OR (IsInt) THEN Mark('no array here!') ELSE BEGIN
	    	      Get(sym); expression(x);
	    	      IF (x.mode <> g_Const) OR (x.a < 0) THEN Mark('bad index');
	    	      IF sym = of_ THEN Get(sym) ELSE Mark('OF?');
	    	      TType(tp); NEW(typet); typet^.form := g_Array; typet^.base := tp;
	    	      typet^.len := (x.a MOD maxint); typet^.size := typet^.len * tp^.size; end;
	    	    END;
	    record_: BEGIN
	    	       Get(sym); NEW(typet); typet^.form := g_Record; typet^.size := 0; OpenScope;
	    	       ok := FALSE;
	    	       REPEAT
	    	         IF sym = ident_ THEN BEGIN
	    	           IdentList(g_Fld, first); TType(tp); obj := first;
	    	           WHILE obj <> guard DO BEGIN
	    	             obj^.o_type := tp; obj^.val := typet^.size; 
	    	             typet^.size := typet^.size + obj^.o_type^.size;;
	    	             //obj^.o_type^.size := obj^.o_type^.size + 2; 
	    	             obj := obj^.next;
	    	           END
	    	         END ;
	    	         IF sym = semicolon_ THEN Get(sym)
	    	         ELSE IF sym = ident_ THEN Mark('; ?')
	    	           ELSE OK := TRUE;
	    	       Until ok;
	    	       typet^.fields := topScope^.next; CloseScope;
	    	       IF sym = end_ THEN Get(sym) ELSE Mark('END?'); END  
	    ELSE Mark('ident?');
	  end;
	END {TType};

	PROCEDURE declarations(VAR varsize: LONGINT);
	  VAR obj, first: g_Object; ok: BOOLEAN;
	    x: g_Item; tp: g_Type; //L: LONGINT;
	    Nequ: Pequ;
 	BEGIN(*sync*)
 	  WHILE sym = semicolon_ DO Get(sym);
	  IF (sym < sfr_) AND (sym <> end_) THEN BEGIN
	    Mark('declaration?');
	    REPEAT Get(sym) UNTIL ((sym >= sfr_) OR (sym = end_));
	  END;
	  ok := FALSE;
	  Repeat
	    WHILE sym = semicolon_ do Get(sym);
	    IF sym = sfr_ THEN BEGIN
	      Get(sym);
	      WHILE sym = ident_ DO BEGIN
	        NewObj(obj, g_SFR); Get(sym); 
	        obj^.o_type := sfrtype; obj^.lev := curlev;
	        { TODO equ ber = einfhren, z.B.: LEDs = P4; Done!}
	        IF sym = eql_ THEN BEGIN
	          Get(sym);
	          IF sym = ident_ THEN BEGIN
	            New(Nequ); Nequ^.name := obj^.name;
	            Nequ^.origin := id; Nequ^.next := Lequ;
	            Lequ := Nequ; Get(sym);
	          END
	          ELSE Mark('SFR-declaration?');
	        END;
	        IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?');
	        WHILE sym = semicolon_ do Get(sym);
	      END;
	    END;
	    IF sym = const_ THEN BEGIN
	      Get(sym);
	      WHILE sym = ident_ DO BEGIN
	        NewObj(obj, g_Const); Get(sym);
	        IF sym = eql_ THEN Get(sym) ELSE Mark('=?');
	        expression(x);
	        IF x.mode = g_Const THEN BEGIN
	          obj^.val := x.a; obj^.o_type := x.i_type; obj^.lev := curlev; END
	        ELSE Mark('expression not constant');
                IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?');
                WHILE sym = semicolon_ do Get(sym);
              END;
            END;
            IF sym = type_ THEN 
              IF smallproc THEN Mark('??')
              ELSE BEGIN
                Get(sym);
                WHILE sym = ident_ DO BEGIN
                  NewObj(obj, g_Typ); Get(sym);
                  IF sym = eql_ THEN Get(sym) ELSE Mark('=?'); 
                  TType(obj^.o_type);
                  IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?');
                  WHILE sym = semicolon_ do Get(sym);
                END;
              END;
            IF sym = var_ THEN BEGIN
              Get(sym);
              WHILE sym = ident_ DO BEGIN
                IdentList(g_Var, first); TType(tp); obj := first;
                IF (smallproc) AND (tp^.form IN [g_Array, g_Record])
		  THEN Mark('no struct. vars in *-procs!')
		ELSE
                WHILE obj <> guard DO BEGIN
                  obj^.o_type := tp; obj^.lev := curlev;
                  varsize := varsize + obj^.o_type^.size; obj^.val := varsize; obj := obj^.next;
                END;
                IF sym = semicolon_ THEN Get(sym) ELSE Mark('; ?');
                WHILE sym = semicolon_ do Get(sym);
              END;
            END ;
            IF (sym >= const_) AND (sym <= var_) THEN Mark('declaration?') ELSE ok := TRUE
	  Until ok;
	END {declarations};
	
	PROCEDURE ProcedureDecl(small: BOOLEAN);
	//CONST marksize = 0;
	VAR proc, obj: g_Object;
	    procid: Ident;
	    locblksize : LONGINT; //, parblksize: LONGINT;
	    Snextreg, Slastreg: INTEGER;
	    ssmallproc: BOOLEAN;
	
		PROCEDURE FPSection;
		  VAR obj, first: g_Object; tp: g_Type; //parsize: LONGINT;
		BEGIN
		  IF sym = var_ THEN BEGIN
		    Get(sym); IdentList(g_Par, first); end
		  ELSE IdentList(g_Var, first);
		  IF sym = ident_ THEN BEGIN
		    find(obj); Get(sym); 
		    IF obj^.class = g_Typ THEN tp := obj^.o_type 
		    ELSE BEGIN
		      Mark('type?'); tp := intType; END;
		  end
		  ELSE BEGIN
		    Mark('ident?'); tp := intType;
		  END;
		  obj := first;
		  WHILE obj <> guard DO BEGIN
		    IF (tp^.form >= g_Array) AND (obj^.class <> g_Par) THEN 
		      Mark('no structered valueparameters');
		    obj^.o_type := tp; //parblksize := parblksize + parsize;
		    obj := obj^.next;
		  END;
		END {FPSection};
		
	BEGIN (* ProcedureDecl *) Get(sym);
	  IF sym = ident_ THEN BEGIN
	    snextreg := nextreg; slastreg := lastreg;
	    ssmallproc := smallproc;
	    IsProc := TRUE;
	    IF small THEN 
	      smallproc := TRUE
	    ELSE smallproc := FALSE;
	    Get(sym);
	    IF (id = 'ISR') THEN BEGIN
	      IF curlev = 0 THEN BEGIN
	        IsInt := TRUE; smallproc := TRUE;
	        nextreg := 0; lastreg := 15;
	        IF sym = lbrak_ THEN BEGIN
	          Get(sym);
	          IF sym = number_ THEN BEGIN
	            procid := Vektoren[val].name;
	            id := procid;
	            Vektoren[val].used := TRUE;
	            NewObj(proc, g_Int); //parblksize := marksize;
	            IncLevel(2); OpenScope; proc^.val := -1;
	            Get(sym);
	            If sym = rbrak_ THEN Get(sym) ELSE Mark(']?');
	          END
	          ELSE Mark('interrupt-vector?');
	        END
	        ELSE Mark('[?');
	      END
	      ELSE Mark('interrupt-level?')
	    END
	    ELSE BEGIN
	      nextreg := 1; lastreg := 15;
	      procid := id; 
	      IF smallproc THEN NewObj(proc, g_ProcR)
	      ELSE NewObj(proc, g_Proc); //Get(sym); 
	      IncLevel(2); OpenScope; proc^.val := -1;
	      IF sym = lparen_ THEN BEGIN
	        Get(sym);
	        IF sym = rparen_ THEN Get(sym)
	        ELSE BEGIN
	          FPSection;
	          WHILE sym = semicolon_ DO BEGIN
	            Get(sym); 
	            IF sym <> rparen_ THEN FPSection
	            ELSE Mark('???');
	          END ;
	          IF sym = rparen_ THEN Get(sym) ELSE Mark(')?');
	        END;
	      END;
	    END;
	    obj := topScope^.next; 
	    locblksize := 0;
	    WHILE obj <> guard DO BEGIN
	      obj^.lev := curlev;
	      IF obj^.o_type^.form IN [g_array, g_record] THEN
	        locblksize := locblksize + WordSize
	      ELSE
	        locblksize := locblksize + obj^.o_type^.size;
	      obj^.val := locblksize; obj^.IsPara := TRUE;
	      obj := obj^.next;
	    END;
	    proc^.dsc := topScope^.next;
	    IF sym = semicolon_ THEN Get(sym) ELSE Mark(';???');
	    declarations(locblksize);
	    WHILE (sym = procedure_) OR (sym = procedureR_) DO BEGIN
	      IF sym = procedureR_ THEN ProcedureDecl(TRUE)
	      ELSE ProcedureDecl(FALSE);
	      IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?');
	    END;
	    SetProcLabel(procid, IsInt);
	    proc^.val := locblksize;
	    IF NOT IsInt THEN BEGIN
	      IF smallproc THEN BEGIN
	        IF locblksize <= maxvarcnt THEN
	          lastreg := 15 - (locblksize DIV 2)
	        ELSE Mark('too many variables for procedure'); END
	      ELSE nextreg := 2;
	    END
	    ELSE BEGIN { Interrupt }
	      IF locblksize <= maxvarcnt THEN
	        lastreg := 15 - (locblksize DIV 2)
	      ELSE Mark('too many variables for interrupt');
	    END;
	    IF sym = begin_ THEN BEGIN
	      Get(sym); StatSequence; END;
	    IF sym = end_ THEN Get(sym) ELSE Mark('END?');
	    IF sym = ident_ THEN BEGIN
	      IF IsInt THEN BEGIN
	        IF id <>'ISR' THEN Mark('no match');
	      END
	      ELSE IF procid <> id THEN Mark('no match');
	      Get(sym);
	    END ;
	    nextreg := snextreg; lastreg := slastreg; smallproc := ssmallproc;
	    Return(proc^.name, IsInt); CloseScope; IncLevel(-2);
	    IsInt := FALSE; IsProc := FALSE;
	  END;
	END {ProcedureDecl};

	PROCEDURE Module;
	  VAR modid: Ident; varsize: LONGINT;
	BEGIN Writeln('   compiling ');
	  if sym = module_ then begin
	    Get(sym); Open; OpenScope; varsize := 0;
	    IF sym = ident_ THEN BEGIN
	      modid := id; Get(sym); end
	    ELSE Mark('ident?');
	    IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?');
	    { Hier msste auf "IMPORT" berprft werden! TODO 
	      Nach entsprechender Dekodierung msste sich MODUKE rekursiv aufrufen. }
	    declarations(varsize);
	    WHILE (sym = procedure_) OR (sym = procedureR_) DO BEGIN
	    	IF sym = procedureR_ THEN ProcedureDecl(TRUE)
	    	ELSE ProcedureDecl(FALSE);
	    	IF sym = semicolon_ THEN Get(sym) ELSE Mark(';?'); 
	    	WHILE sym = semicolon_ DO Get(sym);
	    END;
	    Init_Sproc(varsize);
	    IF sym = begin_ THEN BEGIN 
	      SetProcLabel('mainstart', FALSE);
	      Get(sym); StatSequence; END ;
	    IF sym = end_ THEN Get(sym) ELSE Mark('END?'); 
	    IF sym = ident_ THEN BEGIN
	      IF modid <> id THEN Mark('no match');
	        Get(sym); 
	    END
	    ELSE Mark('ident?');
	    IF sym <> period_ THEN Mark('. ?');
	    CloseScope; {EndMod; s. TODO oben }
	    IF not error THEN BEGIN
	      Writeln('code generated');
	     end;
	  end
	  else Mark('MODULE?');
	END {Module};


	PROCEDURE Compile(fname: String);
	BEGIN 
	  writeln('Oberon compiler for C16x, version ', verstr);
          Init(fname,0); 
          Open_Gen;
          Get(sym); Module;
          End_Mod;
	  closefiles; 
	  writeln;
	END {Compile};



	PROCEDURE enter(cl: INTEGER; n: LONGINT; name: Ident; p_type: g_Type);
	  VAR obj: g_Object;
	BEGIN NEW(obj);
	  obj^.class := cl; obj^.val := n; obj^.name := name; obj^.o_type := p_type; obj^.dsc := NIL;
	  obj^.next := topScope^.next; topScope^.next := obj;
	END {enter}; 

initialization

  NEW(guard); guard^.class := g_Var; guard^.o_type := intType; guard^.val := 0;
  topScope := NIL; OpenScope;
  enter(g_Typ, 1, 'BOOLEAN', boolType);
  enter(g_Typ, 2, 'INTEGER', intType);
  enter(g_Typ, 1, 'SFR', sfrType);
  enter(g_Typ, 1, 'STRING', stringType);
  enter(g_Const, 1, 'TRUE', boolType);
  enter(g_Const, 0, 'FALSE', boolType);
  enter(g_SProc, 1, 'Read', NIL);
  enter(g_SProc, 2, 'Write', NIL);
  enter(g_SProc, 4, 'WriteLn', NIL);
  enter(g_SProc, 5, 'Sleep', NIL);
  universe := topScope;
  ProcPar := FALSE;

end.